#include <lua.h>
#include <lauxlib.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>

/* evaluates to the triangular number */
#define trinum(x) ((x)*((x)+1)/2)

/* returns a random number greater than 0.0 and smaller than 1.0 */
double multivar_random01(lua_State *L, FILE *dev) {
  uint64_t n;
  double f;
  while (1) {
    /* get random integer by reading from random byte generator */
    if (fread(&n, sizeof(n), 1, dev) != 1)
      luaL_error(L, "I/O error when reading from random device");
    /* convert result into float between 0.0 and 1.0 and avoid both 0.0 and 1.0
       as result */
    f = (n + 1.0) / 18446744073709551616.0;
    if (f < 1.0) return f;
  }
}

/* returns two independent, (0,1) normally distributed numbers */
void multivar_stddist(lua_State *L, FILE *dev, double *x, double *y) {
  /* use box-muller transform of two equally distributed random numbers */
  double r = sqrt(-2.0 * log(multivar_random01(L, dev)));
  double p = (2.0 * M_PI) * multivar_random01(L, dev);
  *x = cos(p) * r;
  *y = sin(p) * r;
}

/* returns an equally distributed random number between 0.0 and 1.0 if no
   arguments are supplied, or a random vector according to a multivariate
   normal distribution with a given covariance (lower triangular matrix as
   first argument) and a given mean (optional vector as second argument) */
int multivar(lua_State *L) {
  lua_Integer dim, i, j, k;
  lua_Number *matrix, *vector, sum;
  /* get number of arguments passed to function */
  int argc = lua_gettop(L);
  /* get pointer to FILE handle connected with random byte generator */
  FILE *dev = lua_touserdata(L, lua_upvalueindex(1));
  /* return equally distributed random number if no arguments have been
     passed */
  if (argc == 0) {
    lua_pushnumber(L, multivar_random01(L, dev));
    return 1;
  }
  /* check argument types */
  luaL_checktype(L, 1, LUA_TTABLE);
  if (argc >= 2) luaL_checktype(L, 2, LUA_TTABLE);
  /* determine dimension baseds on length of first argument */
  lua_len(L, 1);
  dim = lua_tointeger(L, -1);
  lua_pop(L, 1);
  /* return empty table if dimension is not positive */
  if (dim <= 0) {
    lua_createtable(L, 0, 0);
    return 1;
  }
  /* allocate triangular matrix */
  if (dim > SIZE_MAX / sizeof(lua_Number) / dim)
    luaL_error(L, "Memory allocation error");
  matrix = lua_newuserdata(L, trinum(dim) * sizeof(lua_Number));
  /* convert nested Lua tables containing the lower triangle of the covariance
     matrix into its Cholesky decomposition (represented as triangular matrix
     in C) */
  for (i=0; i<dim; i++) {
    lua_pushinteger(L, i+1);
    lua_gettable(L, 1);
    luaL_argcheck(
      L, lua_type(L, -1) == LUA_TTABLE, 1,
      "table in table expected"
    );
    for (j=0; j<=i; j++) {
      lua_pushinteger(L, j+1);
      lua_gettable(L, -2);
      luaL_argcheck(
        L, lua_type(L, -1) == LUA_TNUMBER, 1,
        "number in table in table expected"
      );
      sum = lua_tonumber(L, -1);
      lua_pop(L, 1);
      for (k=0; k<j; k++) sum -= matrix[trinum(i)+k] * matrix[trinum(j)+k];
      matrix[trinum(i)+j] = (
        (i != j) ? (sum / matrix[trinum(j)+j]) :
        (sum > 0) ? sqrt(sum) : 0
      );
    }
    lua_pop(L, 1);
  }
  /* allocate and initialize C array with independent, normally distributed
     values */
  vector = lua_newuserdata(L, dim * sizeof(lua_Number));
  for (i=0; i<dim-1; i+=2) {
    multivar_stddist(L, dev, vector+i, vector+i+1);
  }
  if (i<dim) multivar_stddist(L, dev, vector+i, vector+i);
  /* multiply Cholesky decomposition matrix with normally distributed random
     vector and, if applicable, add vector with the desired mean */
  lua_createtable(L, dim, 0);
  for (i=0; i<dim; i++) {
    if (argc < 2) sum = 0;
    else {
      lua_pushinteger(L, i+1);
      lua_gettable(L, 2);
      luaL_argcheck(
        L, lua_type(L, -1) == LUA_TNUMBER, 2,
        "number in table expected"
      );
      sum = lua_tonumber(L, -1);
      lua_pop(L, 1);
    }
    for (j=0; j<=i; j++) sum += vector[j] * matrix[trinum(i)+j];
    lua_pushnumber(L, sum);
    lua_rawseti(L, -2, i+1);
  }
  /* return result (matrix vector multiplication with optional addition) */
  return 1;
}

/* initializes library by opening /dev/urandom and returning a Lua function */
int luaopen_multivar(lua_State *L) {
  FILE *dev = fopen("/dev/urandom", "r");
  if (!dev) return luaL_error(L, "Could not open /dev/urandom");
  lua_pushlightuserdata(L, dev);
  lua_pushcclosure(L, multivar, 1);
  return 1;
}
